; 

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

align 16

utoa proc src:DWORD, lpdest:DWORD

  ; --------------------------------------------------------
  ; This is Ray's original algo with the stack frame removed
  ; --------------------------------------------------------
      push ebx
      push edi

      mov   edi,[esp+8][8]    ;lpdest        ;buffer address
      mov   eax,[esp+4][8]    ;src
      add   edi,10            ;=>11th byte of the buffer
      push  edi
      mov   ecx,0CCCCCCCDh    ;"magic number" multiplier for division by 10
      mov   ebx,10
      mov   byte ptr[edi],0   ;string terminating 0
   @@:
      mul   ecx               ;multiply by magic number
      shrd  eax,edx,3         ;binary fractional "remainder" back in EAX
      shr   edx,3             ;EDX = quotient
      inc   eax               ;precaution against occasional "underflow"
      push  edx               ;save current quotient for next division
      mul   ebx               ;x10 gets "decimal" remainder into EDX
      dec   edi               ;back to previous digit in buffer
      add   dl,30h            ;convert remainder to ascii
      mov   [edi],dl          ;insert it into buffer
      pop   eax               ;retrieve current quotient
      test  eax,eax           ;test if done
      jnz   @B                ;continue if not done

      pop   ecx
      mov   eax,edi           ;EAX = address of first character
      sub   ecx,edi           ;ECX = number of characters

      pop edi
      pop ebx

      ret   8

utoa endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

; 

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

    align 16

utoa_ex proc uvar:DWORD,pbuffer:DWORD

  ; --------------------------------------------------------------------------------
  ; this algorithm was written by Paul Dixon and has been converted to MASM notation
  ; --------------------------------------------------------------------------------

    mov eax, [esp+4]                ; uvar      : unsigned variable to convert
    mov ecx, [esp+8]                ; pbuffer   : pointer to result buffer

    push esi
    push edi

    jmp udword

  align 4
  chartab:
    dd "00","10","20","30","40","50","60","70","80","90"
    dd "01","11","21","31","41","51","61","71","81","91"
    dd "02","12","22","32","42","52","62","72","82","92"
    dd "03","13","23","33","43","53","63","73","83","93"
    dd "04","14","24","34","44","54","64","74","84","94"
    dd "05","15","25","35","45","55","65","75","85","95"
    dd "06","16","26","36","46","56","66","76","86","96"
    dd "07","17","27","37","47","57","67","77","87","97"
    dd "08","18","28","38","48","58","68","78","88","98"
    dd "09","19","29","39","49","59","69","79","89","99"

  udword:
    mov esi, ecx                    ; get pointer to answer
    mov edi, eax                    ; save a copy of the number

    mov edx, 0D1B71759h             ; =2^45\10000    13 bit extra shift
    mul edx                         ; gives 6 high digits in edx

    mov eax, 68DB9h                 ; =2^32\10000+1

    shr edx, 13                     ; correct for multiplier offset used to give better accuracy
    jz short skiphighdigits         ; if zero then don't need to process the top 6 digits

    mov ecx, edx                    ; get a copy of high digits
    imul ecx, 10000                 ; scale up high digits
    sub edi, ecx                    ; subtract high digits from original. EDI now = lower 4 digits

    mul edx                         ; get first 2 digits in edx
    mov ecx, 100                    ; load ready for later

    jnc short next1                 ; if zero, supress them by ignoring
    cmp edx, 9                      ; 1 digit or 2?
    ja   ZeroSupressed              ; 2 digits, just continue with pairs of digits to the end

    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dh                   ; but only write the 1 we need, supress the leading zero
    inc esi                         ; update pointer by 1
    jmp  ZS1                        ; continue with pairs of digits to the end

  align 16
  next1:
    mul ecx                         ; get next 2 digits
    jnc short next2                 ; if zero, supress them by ignoring
    cmp edx, 9                      ; 1 digit or 2?
    ja   ZS1a                       ; 2 digits, just continue with pairs of digits to the end

    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dh                   ; but only write the 1 we need, supress the leading zero
    add esi, 1                      ; update pointer by 1
    jmp  ZS2                        ; continue with pairs of digits to the end

  align 16
  next2:
    mul ecx                         ; get next 2 digits
    jnc short next3                 ; if zero, supress them by ignoring
    cmp edx, 9                      ; 1 digit or 2?
    ja   ZS2a                       ; 2 digits, just continue with pairs of digits to the end

    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dh                   ; but only write the 1 we need, supress the leading zero
    add esi, 1                      ; update pointer by 1
    jmp  ZS3                        ; continue with pairs of digits to the end

  align 16
  next3:

  skiphighdigits:
    mov eax, edi                    ; get lower 4 digits
    mov ecx, 100

    mov edx, 28F5C29h               ; 2^32\100 +1
    mul edx
    jnc short next4                 ; if zero, supress them by ignoring
    cmp edx, 9                      ; 1 digit or 2?
    ja  short ZS3a                  ; 2 digits, just continue with pairs of digits to the end

    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dh                   ; but only write the 1 we need, supress the leading zero
    inc esi                         ; update pointer by 1
    jmp short  ZS4                  ; continue with pairs of digits to the end

  align 16
  next4:
    mul ecx                         ; this is the last pair so don; t supress a single zero
    cmp edx, 9                      ; 1 digit or 2?
    ja  short ZS4a                  ; 2 digits, just continue with pairs of digits to the end

    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dh                   ; but only write the 1 we need, supress the leading zero
    mov byte ptr [esi+1], 0         ; zero terminate string

    pop edi
    pop esi
    ret 8

  align 16
  ZeroSupressed:
    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dx
    add esi, 2                      ; write them to answer

  ZS1:
    mul ecx                         ; get next 2 digits
  ZS1a:
    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dx                   ; write them to answer
    add esi, 2

  ZS2:
    mul ecx                         ; get next 2 digits
  ZS2a:
    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dx                   ; write them to answer
    add esi, 2

  ZS3:
    mov eax, edi                    ; get lower 4 digits
    mov edx, 28F5C29h               ; 2^32\100 +1
    mul edx                         ; edx= top pair
  ZS3a:
    mov edx, chartab[edx*4]         ; look up 2 digits
    mov [esi], dx                   ; write to answer
    add esi, 2                      ; update pointer

  ZS4:
    mul ecx                         ; get final 2 digits
  ZS4a:
    mov edx, chartab[edx*4]         ; look them up
    mov [esi], dx                   ; write to answer

    mov byte ptr [esi+2], 0         ; zero terminate string

  sdwordend:

    pop edi
    pop esi

    ret 8

utoa_ex endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

;  

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

    align 16

utoa2 PROC, value:DWORD, szbuf:DWORD

  ; ------------------------------------------------
  ; This is brethren's conversion with the buffer
  ; length calculated from the value to be converted
  ; ------------------------------------------------
    push ebx
    push esi

    mov eax, [esp+4][8]       ; value
    mov esi, [esp+8][8]       ; szbuf

    cmp eax, 1000000000
    jb lb9
    add esi, 10

  nxt:
    mov ebx, 0CCCCCCCDh
    mov BYTE PTR [esi], 0
    sub esi, 1

  align 4
  stlp:
    mov ecx, eax
    mul ebx
    shr edx, 3
    mov eax, edx
    lea edx, [edx+edx*4]
    lea edx, [edx+edx-48]
    sub ecx, edx
    mov [esi], cl
    sub esi, 1
    test eax, eax
    jbe lpqt

    mov ecx, eax
    mul ebx
    shr edx, 3
    mov eax, edx
    lea edx, [edx+edx*4]
    lea edx, [edx+edx-48]
    sub ecx, edx
    mov [esi], cl
    sub esi, 1
    test eax, eax
    ja stlp

  lpqt:
    pop esi
    pop ebx

    ret 8

  align 4
  lb1:
    add esi, 1
    jmp nxt
  lb2:
    cmp eax, 10
    jb lb1
    add esi, 2
    jmp nxt
  lb3:
    cmp eax, 100
    jb lb2
    add esi, 3
    jmp nxt
  lb4:
    cmp eax, 1000
    jb lb3
    add esi, 4
    jmp nxt
  lb5:
    cmp eax, 10000
    jb lb4
    add esi, 5
    jmp nxt
  lb6:
    cmp eax, 100000
    jb lb5
    add esi, 6
    jmp nxt
  lb7:
    cmp eax, 1000000
    jb lb6
    add esi, 7
    jmp nxt
  lb8:
    cmp eax, 10000000
    jb lb7
    add esi, 8
    jmp nxt
  lb9:
    cmp eax, 100000000
    jb lb8
    add esi, 9
    jmp nxt

utoa2 ENDP

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

;  

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

    align 16

utoa_exLingo proc src:DWORD, lpdest:DWORD

; -----------------------------------------------------------------------------------
; This version was modified with Lingo's suggestion to balance the CALL / RET pairing.
; -----------------------------------------------------------------------------------
  .data
    align 16
    chartabL dw "00","10","20","30","40","50","60","70","80","90"
             dw "01","11","21","31","41","51","61","71","81","91"
             dw "02","12","22","32","42","52","62","72","82","92"
             dw "03","13","23","33","43","53","63","73","83","93"
             dw "04","14","24","34","44","54","64","74","84","94"
             dw "05","15","25","35","45","55","65","75","85","95"
             dw "06","16","26","36","46","56","66","76","86","96"
             dw "07","17","27","37","47","57","67","77","87","97"
             dw "08","18","28","38","48","58","68","78","88","98"
             dw "09","19","29","39","49","59","69","79","89","99"
  .code

    mov eax, [esp+4]            ; src
    mov edx, 0D1B71759h
    mov ecx, [esp+8]            ; lpdest
    mov [esp+4], edi            ; preserve EDI on stack
    mov edi, eax
    mov [esp+8], esi            ; preserve ESI on stack
    mul edx
    shr edx, 13
    mov esi, 100
    je lbl11
    imul eax, edx, 10000
    sub edi, eax
    mov eax, 68DB9h
    mul edx
    jnb lbl9
    cmp edx, 9
    jbe lbl8
    movzx edx, WORD PTR [edx+edx+chartabL]
    add ecx, 8
    mov [ecx-8], edx

  lbl0:
    mul esi

  lbl1:
    movzx edx, WORD PTR [edx+edx+chartabL]
    mov [ecx-6], edx

  lbl2:
    mul esi

  lbl3:
    movzx edx, WORD PTR [edx+edx+chartabL]
    mov [ecx-4], edx

  lbl4:
    mov eax, 28F5C29h
    mul edi

  lbl5:
    movzx edx, WORD PTR [edx+edx+chartabL]
    mov [ecx-2], edx

  lbl6:
    mul esi

  lbl7:
    movzx eax, WORD PTR [edx+edx+chartabL]
    pop edx
    mov [ecx], eax
    pop edi
    lea eax, [ecx+2]
    pop esi

    push edx
    ret

  lbl8:
    add ecx, 7
    add edx, 48
    mov [ecx-7], edx
    jnz lbl0

  lbl9:
    mul esi
    jnb lbl10
    add ecx, 6
    cmp edx, 9
    ja lbl1
    add edx, 48
    sub ecx, 1
    mov [ecx-5], edx
    jnz lbl2

  lbl10:
    mul esi
    jnb lbl11
    add ecx, 4
    cmp edx, 9
    ja lbl3
    add edx, 48
    sub ecx, 1
    mov [ecx-3], edx
    jnz lbl4

  lbl11:
    mov eax, 28F5C29h
    mul edi
    jnb lbl12
    add ecx, 2
    cmp edx, 9
    ja lbl5
    add edx, 48
    sub ecx, 1
    mov [ecx-1], edx
    jnz lbl6

  lbl12:
    mul esi
    cmp edx, 9
    ja lbl7
    lea eax, [edx+48]
    pop edx
    mov [ecx], eax
    pop edi
    lea eax, [ecx+1]
    pop esi

    push edx
    ret

utoa_exLingo endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

;  

OPTION PROLOGUE:NONE
OPTION EPILOGUE:NONE

    align 16

utoa3 proc dwNumber:DWORD, pszString:DWORD

  ; --------------------------------------------
  ; This is a modified version of Ray's algo to
  ; align the result to the buffer start address
  ; --------------------------------------------

    mov edx, [esp+4]

    cmp edx, 9
    ja @F
    mov edx, 1
    jmp nxt

  @@:
    cmp edx, 99
    ja @F
    mov edx, 2
    jmp nxt

  @@:
    cmp edx, 999
    ja @F
    mov edx, 3
    jmp nxt

  @@:
    cmp edx, 9999
    ja @F
    mov edx, 4
    jmp nxt

  @@:
    cmp edx, 99999
    ja @F
    mov edx, 5
    jmp nxt

  @@:
    cmp edx, 999999
    ja @F
    mov edx, 6
    jmp nxt

  @@:
    cmp edx, 9999999
    ja @F
    mov edx, 7
    jmp nxt

  @@:
    cmp edx, 99999999
    ja @F
    mov edx, 8
    jmp nxt

  @@:
    cmp edx, 999999999
    ja @F
    mov edx, 9
    jmp nxt

  @@:
    mov edx, 10

  nxt:
    mov eax, [esp+4]            ; src
    mov ecx, [esp+8]            ; lpdest             ; buffer address

    push ebx
    push esi
    push edi

    mov esi, 0CCCCCCCDh         ; "magic number" multiplier for division by 10
    mov ebx, 10
    add ecx, edx                ; add required buffer length to start address
    mov BYTE PTR [ecx], 0       ; write the terminator
    sub ecx, 1                  ; decrement down to first write position

  @@:
    mul esi                     ; multiply by magic number
    shrd eax, edx, 3            ; binary fractional "remainder" back in EAX
    shr edx, 3                  ; EDX = quotient
    add eax, 1                  ; precaution against occasional "underflow"
    mov edi, edx                ; save current quotient for next division
    mul ebx                     ; x10 gets "decimal" remainder into EDX
    add dl, 30h                 ; convert remainder to ascii
    mov [ecx], dl               ; insert it into buffer
    sub ecx, 1                  ; decrement buffer location
    mov eax, edi                ; retrieve current quotient
    test eax, eax               ; test if done
    jz @F                       ; continue if not done

    mul esi                     ; multiply by magic number
    shrd eax, edx, 3            ; binary fractional "remainder" back in EAX
    shr edx, 3                  ; EDX = quotient
    add eax, 1                  ; precaution against occasional "underflow"
    mov edi, edx                ; save current quotient for next division
    mul ebx                     ; x10 gets "decimal" remainder into EDX
    add dl, 30h                 ; convert remainder to ascii
    mov [ecx], dl               ; insert it into buffer
    sub ecx, 1                  ; decrement buffer location
    mov eax, edi                ; retrieve current quotient
    test eax, eax               ; test if done
    jnz @B                      ; continue if not done


  @@:
    pop edi
    pop esi
    pop ebx

    mov eax, [esp+8]

    ret 8

utoa3 endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef

;  
